bl_info = {
    "name": "Import RBDOOM-3-BFG JSON map format (.json)",
    "author": "Robert Beckebans",
    "version": (0, 8, 1),
    "blender": (2, 83, 1),
    "location": "File > Import > Doom 3 Map (.json)",
    "description": "Imports a Doom 3 map that uses the RBDOOM-3-BFG .json format",
    "warning": "",
    "wiki_url": "",
    "category": "Import-Export",
}

import json, sys, os, bpy, bpy_extras
import mathutils
import time
from decimal import *
from math import *

from bpy.props import (
        StringProperty,
        BoolProperty,
        FloatProperty,
        EnumProperty,
        )
        
from bpy_extras.io_utils import (
        ImportHelper,
        ExportHelper,
        orientation_helper,
        path_reference_mode,
        axis_conversion,
        )

# path helper
def splitall(path):
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts



def get_vector( data, key ):
    v = list( map( Decimal, data[ key ].split() ) )
    vec = mathutils.Vector( ( v[0], v[1], v[2] ) )
    return vec




def import_map( filename ):
    
    # find basepath which is either base or a mod folder
    filename_split = splitall( filename )
    print( "filename = ", filename_split )
    
    #return
    
    basepath = ''
    for part in filename_split:
        
        if part == "maps":
            break
        
        basepath = os.path.join( basepath, part )
    
    print( "basepath = ", basepath )
    
    deffilename = os.path.join( basepath, 'exported', 'entities.json' )
    
    print( "definitions file = ", deffilename )
    
    defs = json.loads( open( deffilename ).read() )

    #if "gravity" in defs["entities"]["aas48"]: 
    #    print( "cool aas48.gravity = {0}".format( data["entities"]["aas48"]["gravity"] ) )
    
    start = time.time() 
    data = json.loads( open( filename ).read() )
    end = time.time()
    print( "loading {0} took {1} seconds".format( filename, ( end - start ) ) )
    
    numEntities = len( data["entities"] )
    for entNum in range( numEntities ):
        
        ent = data["entities"][entNum]
     
        origin = ( 0, 0, 0 )
        entname = "worldspawn"
        classname = ent["classname"]
        
        if "origin" in ent:
            origin = get_vector( ent, "origin" )
            
        if "name" in ent:
            entname = ent["name"]
            
        print( "creating entity {0} of {1}: {2}".format( entNum, numEntities, entname ) )
            
        #group = bpy.data.groups.new( entname )
        #group = bpy.ops.object.empty_add()
        
        entObj = None
        
        if "light_radius" in ent:
            # create new lamp datablock
            lightData = bpy.data.lights.new( name = entname, type='POINT')
            
            """
            FIXME FIXME
            """
            radius = get_vector( ent, "light_radius" )
            print( radius.magnitude )
            lightData.energy = radius.magnitude * 750.0
            lightData.use_nodes = True

            # create new object with our lamp datablock
            entObj = bpy.data.objects.new( name = entname, object_data=lightData )

            # link light object
            bpy.context.collection.objects.link( entObj )

            # place light to a specified location
            entObj.location = origin
            entObj.show_name = True

            # and finally select it make active
            #entObj.select = True
            #bpy.context.collection.objects.active = entObj
            #bpy.context.view_layer.objects.active = entObj
            
        
        else:
        
            bpy.ops.object.add(
                type='EMPTY', 
                enter_editmode=False,
                location=origin)
        
            entObj = bpy.context.object
            entObj.name = entname
            entObj.show_name = True
            #entObj.select = True
            
        #bpy.context.scene.objects.active = entObj
        
        for key in ent:
            
            if key != "primitives":
                
                value = ent[ key ]
                
                if key in defs["entities"][classname]:
                    #print( "game property {0}.{1} = {2}, default = {3}".format( classname, key, value, defs["entities"][classname][key] ) )
                    #group[ key ] = ent[ key ]
                    
                    #bpy.ops.object.game_property_new( type = 'STRING', name = key )
                    
                    if key != "entity" and key != "name" and key != "origin" and key != "angle":
                        
                        if key == "nodiffuse":
                            entObj[ key ]= int( value )
                        else:
                            entObj[ key ] = value
                else:
                    #if key != "entity" and key != "name" and key != "origin" and key != "angle" and key != "rotation":
                    #    print( "unknown game property {0}.{1} = {2}".format( classname, key, value ) )
                    
                    #bpy.ops.object.game_property_new( type = 'STRING', name = key )
                    #bpy.context.object.value = value
                    
                    if key != "entity" and key != "name" and key != "origin" and key != "angle":
                        entObj[ key ] = value
        
        #print( "origin =", origin )
     
        if "primitives" in ent:
            for i in range( len( ent["primitives"] ) ):
                
                prim = ent["primitives"][i]
                #print( "numverts =", len( prim["verts"] ) )
                
                primName = "{0}.prim{1}".format( entname, i )
                
                mesh = bpy.data.meshes.new( primName )
                meshObj = bpy.data.objects.new( primName, mesh )
                #meshObj.location = origin
                meshObj.show_name = True
     
                # link object to scene and make active
                bpy.context.collection.objects.link( meshObj )
                #scn.objects.active = meshObj
                #meshObj.select = True
        
                #bpy.ops.object.group_link( group = group.name )
                meshObj.parent = entObj
     
                
                # build basic mesh
                verts = []
                for v in prim["verts"]:
                    verts.append( ( v["xyz"][0], v["xyz"][1], v["xyz"][2] ) )
                    
                faces = []
                for f in prim["polygons"]:
                    faces.append( f["indices"] )
                    
                mesh.from_pydata(verts, [], faces)
                
                # assign UVs
                mesh.uv_layers.new( do_init = False, name="UVMap")
                for i in range( len( prim["verts"] ) ):
                    v = prim["verts"][i]
                    mesh.uv_layers[0].data[i].uv = ( v["st"][0], 1.0 - v["st"][1] )
                    
                # collect materials
                materials = []
                for f in prim["polygons"]:
                    mat = f["material"]
                    if not mat in materials:
                        materials.append( mat )
                
                # create blender materials 
                for mat in materials:
                    #print( mat )
                    
                    bmat = None
                    if mat in bpy.data.materials:
                        bmat = bpy.data.materials[ mat ]
                    else:
                        bmat = bpy.data.materials.new( mat )
                    
                    mesh.materials.append( bmat )
                    
                # create blender groups for special materials like textures/common/clip or visportals
                groupDict = [
                    ( "textures/common/caulk", "caulk" ),
                    ( "textures/common/nodraw", "nodraw" ),
                    ( "textures/common/nodrawsolid", "nodrawsolid" ),
                    
                    ( "textures/common/collision", "collision" ),
                    ( "textures/common/clip", "clip" ),
                    ( "textures/common/player_clip", "player_clip" ),
                    ( "textures/common/player_clip", "clip" ),
                    ( "textures/common/full_clip", "full_clip" ),
                    ( "textures/common/full_clip", "clip" ),
                    ( "textures/common/monster_clip", "monster_clip" ),
                    ( "textures/common/monster_clip", "clip" ),
                    ( "textures/common/ik_clip", "ik_clip" ),
                    ( "textures/common/ik_clip", "clip" ),
                    ( "textures/common/movable_clip", "movable_clip" ),
                    ( "textures/common/movable_clip", "clip" ),
                    ( "textures/common/clip_plusmovables", "clip_plusmovables" ),
                    ( "textures/common/clip_plusmovables", "clip" ),
                    
                    ( "textures/common/cushion", "cushion" ),
                    ( "textures/common/slick", "slick" ),
                    ( "textures/common/noimpact", "noimpact" ),
                    ( "textures/common/nodrop", "nodrop" ),
                    ( "textures/common/ladder", "ladder" ),
                    
                    ( "textures/common/mirror", "mirror" ),
                    
                    ( "textures/common/shadow", "shadow" ),
                    ( "textures/common/shadow2", "shadow2" ),
                    
                    ( "textures/common/trigger", "trigger" ),
                    ( "textures/common/trigmulti", "trigger" ),
                    ( "textures/common/trigonce", "trigger" ),
                    ( "textures/common/trigtimer", "trigger" ),
                    ( "textures/common/trigrelay", "trigger" ),
                    ( "textures/common/trighurt", "trigger" ),
                    ( "textures/common/trigfade", "trigger" ),
                    ( "textures/common/triggui", "trigger" ),
                    ( "textures/common/trigentityname", "trigger" ),
                    ( "textures/common/trigentityname_once", "trigger" ),
                    
                    ( "textures/editor/entitygui", "gui" ),
                    ( "textures/editor/entitygui2", "gui" ),
                    ( "textures/editor/entitygui3", "gui" ),
                    ( "textures/editor/pda_gui", "gui" ),
                    
                    ( "textures/editor/aassolid", "aassolid" ),
                    ( "textures/editor/aassolid", "aas" ),
                    ( "textures/editor/aasobstacle", "aasobstacle" ),
                    ( "textures/editor/aassolid", "aas" ),
                    
                    ( "textures/editor/visportal", "visportal" )
                ]
                
                for mat in materials:
                    
                    for ( tex, groupName ) in groupDict:
                        if mat == tex:
                            group = None
                            if groupName in bpy.data.collections:
                                collection = bpy.data.collections[ groupName ]
                                collection.objects.link( meshObj )
                            else:
                                bpy.ops.collection.create( name = groupName )
                            
                
                    
                for i in range( len( prim["polygons"] ) ):
                    f = prim["polygons"][i]
                    mat = f["material"]
                    mesh.polygons[i].material_index = materials.index( mat )
                    #print( "polygon {0} material_index = {1}".format( i, me.polygons[i].material_index ) )
                
                mesh.update()
                
                #meshObj.select = False
            
        # deselect so we don't add objects into the wrong groups
        #entObj.select = False



class ImportMap( bpy.types.Operator, ImportHelper ):
    """Load a JSON file"""
    bl_idname = "import_scene.rbdoom_map_json"
    bl_label = "Import RBDOOM-3-BFG JSON map"
    bl_options = {'UNDO', 'PRESET'}

    directory = StringProperty()

    filename_ext = ".json"
    filter_glob = StringProperty( default="*.json", options = { 'HIDDEN' } )

    def execute(self, context):
        import_map( self.filepath )
        
        return { 'FINISHED' }
        

def menu_func_import(self, context):
    self.layout.operator( ImportMap.bl_idname, text="RBDOOM-3-BFG Map (.json)")


classes = (
    ImportMap,
)

def make_annotations(cls):
    """Converts class fields to annotations if running with Blender 2.8"""
    if bpy.app.version < (2, 80):
        return cls
    bl_props = {k: v for k, v in cls.__dict__.items() if isinstance(v, tuple)}
    if bl_props:
        if '__annotations__' not in cls.__dict__:
            setattr(cls, '__annotations__', {})
        annotations = cls.__dict__['__annotations__']
        for k, v in bl_props.items():
            annotations[k] = v
            delattr(cls, k)
    return cls





def register():
    for cls in classes:
        make_annotations( cls ) 
        bpy.utils.register_class( cls )

    bpy.types.TOPBAR_MT_file_import.append( menu_func_import )


def unregister():
    for cls in classes:
        bpy.utils.unregister_class( cls )

    bpy.types.INFO_MT_file_import.remove( menu_func_import )

if __name__ == "__main__":
    register()
    
    # test for fast iteration
    #filename = 'C:/Projects/RBDOOM-3-BFG/base/maps/testmaps/test_box.json'
    #filename = 'C:/Projects/RBDOOM-3-BFG/base/maps/game/pdas.json'
    #import_map( filename )


bl_info = {
    "name": "Export RBDOOM-3-BFG JSON map format (.json)",
    "author": "Robert Beckebans",
    "version": (2, 7),
    "blender": (2, 65, 4),
    "location": "File > Export > Doom 3 Map (.json)",
    "description": "Exports a Doom 3 map that uses the RBDOOM-3-BFG .json format",
    "warning": "",
    "wiki_url": "",
    "category": "Import-Export",
}

import json, sys, bpy, os
import mathutils
import time
from decimal import *
from math import *

from bpy.props import (
        StringProperty,
        BoolProperty,
        FloatProperty,
        EnumProperty,
        )
from bpy_extras.io_utils import (
        ImportHelper,
        ExportHelper,
        orientation_helper_factory,
        path_reference_mode,
        axis_conversion,
        )

def object_children_to_primitives( obj ):
    has_primitives = False
    
    for c in obj.children:
        if c.type == 'MESH':
            has_primitives = True
            break
        
    primitives = []
    if has_primitives:
        for c in obj.children:
            
            if c.type == 'MESH':
                primitive = {}
                verts = []
                
                if len( c.data.uv_layers ) > 0:
                    
                    for i in range( len( c.data.vertices ) ):
                        v = c.data.vertices[i]
                        st = c.data.uv_layers[0].data[i]
                        verts.append( { 'xyz': [ v.co[0], v.co[1], v.co[2] ], 'st': [ st.uv[0], 1.0 - st.uv[1] ], 'normal': [ v.normal[0], v.normal[1], v.normal[2] ] } )
                        
                else:
                    for v in c.data.vertices:
                        verts.append( { 'xyz': [ v.co[0], v.co[1], v.co[2] ], 'normal': [ v.normal[0], v.normal[1], v.normal[2] ] } )    
                    
                primitive['verts'] = verts
                primitives.append( primitive )
                
                polygons = []
                
                for p in c.data.polygons:
                    #print ( p.vertices )
                    indices = []
                    
                    for i in p.vertices:
                        indices.append( i )
                    
                    #for i in range( len( p.vertices ) ):
                    #    indices.append( p.vertices[i] )
                    
                    polygons.append( { 'material': c.data.materials[p.material_index].name, 'indices': indices } )
                    
                primitive['polygons'] = polygons
                
    return primitives

def object_to_entity( obj ):
    #return { 'name': obj.name }
    #return { { 'name': key } for key in obj.keys() }
    #return { "children":[{'name':key,"size":value} for key,value in obj.items()]}
    #return {'name':key, "size":value} for key,value in obj.items()
    #return json.dumps( obj.items(), indent = 4  )
    #return { "items": obj.items() }
    
    #return { "children":[{'name':key,"size":value} for key,value in obj.items()]}
    #return { obj.items() }
    #return sample
    
    #dict = {}
    #for key,value in obj.items():
    #   dict.insert( key, value )
    
    print( "creating entity {0} of type {1}".format( obj.name, obj.type ) )
    
    # make shallow copy without arrays of entries
    #ent = dict( obj )
    #ent = dict()
    
    ignore_keys = [ '_RNA_UI', 'cycles_visibility', 'cycles' ]
    
    ent = {}
    for key,value in obj.items():# if key not in ignore_keys:
        
       if key not in ignore_keys:
            ent[ key] = value
            print( "entity {0} key {1} has value {2}".format( obj.name, key, value ) )
    
    
    # TODO quakeify name if necessary
    ent['name'] = obj.name
    ent['origin'] = "%f %f %f" % ( obj.location[0], obj.location[1], obj.location[2] )
    #ent['blender_object_type'] = obj.type
    
    if obj.type == 'LAMP':
        
        data = obj.data
        color = data.color
        ent['_color'] = "%d %d %d" % ( color[0], color[1], color[2] )
        
        distance = data.distance
        ent['light_radius'] = "%f %f %f" % ( distance, distance, distance )
        
    if obj.type == 'EMPTY':
        primitives = object_children_to_primitives( obj )
        
        if len( primitives ) > 0:
            ent['primitives'] = primitives
   
    return ent
    
    
    
def export_map( filename ):
    ignore_types = ['CAMERA','MESH','ARMATURE']

    entities = [ obj for obj in bpy.context.scene.objects if obj.type not in ignore_types ]

    data = {
        "version": 3,
        "authoring_tool": "Blender " + bpy.app.version_string,
        "entities": [ object_to_entity( obj ) for obj in entities ] }

    j = json.dumps( data, indent = 4 )

    f = open( filename, 'w' )
    #print( j )
    f.write( j )
    f.close()


class ExportMap( bpy.types.Operator, ExportHelper ):
    """Save a JSON file"""
    bl_idname = "export_scene.rbdoom_map_json"
    bl_label = "Export RBDOOM-3-BFG JSON map"
    bl_options = {'UNDO', 'PRESET'}

    directory = StringProperty()

    filename_ext = ".json"
    filter_glob = StringProperty( default="*.json", options = { 'HIDDEN' } )

    def execute(self, context):
        export_map( self.filepath )
        
        return { 'FINISHED' }
        

def menu_func_export(self, context):
    self.layout.operator( ExportMap.bl_idname, text="RBDOOM-3-BFG Map (.json)")


def register():
    bpy.utils.register_module(__name__)

    bpy.types.INFO_MT_file_export.append(menu_func_export)


def unregister():
    bpy.utils.unregister_module(__name__)

    bpy.types.INFO_MT_file_export.remove(menu_func_export)

if __name__ == "__main__":
    register()
    
    """
    ignore_types = ['CAMERA','MESH','ARMATURE']

    entities = [ obj for obj in bpy.context.selected_objects if obj.type not in ignore_types ]

    data = {
        "version": 3,
        "authoring_tool": "Blender " + bpy.app.version_string,
        "entities": [ object_to_entity( obj ) for obj in entities ] }

    j = json.dumps( data, indent = 4 )
    print( j )
    """


bl_info = {
    "name": "Import RBDOOM-3-BFG JSON map format (.json)",
    "author": "Robert Beckebans",
    "version": (2, 7),
    "blender": (2, 65, 4),
    "location": "File > Import > Doom 3 Map (.json)",
    "description": "Imports a Doom 3 map that uses the RBDOOM-3-BFG .json format",
    "warning": "",
    "wiki_url": "",
    "category": "Import-Export",
}

import json, sys, bpy, os
import mathutils
import time
from decimal import *
from math import *

from bpy.props import (
        StringProperty,
        BoolProperty,
        FloatProperty,
        EnumProperty,
        )
from bpy_extras.io_utils import (
        ImportHelper,
        ExportHelper,
        orientation_helper_factory,
        path_reference_mode,
        axis_conversion,
        )

# path helper
def splitall(path):
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts



def get_vector( data, key ):
    v = list( map( Decimal, data[ key ].split() ) )
    vec = mathutils.Vector( ( v[0], v[1], v[2] ) )
    return vec

def import_map( filename ):
    
    # find basepath which is either base or a mod folder
    filename_split = splitall( filename )
    #print( "filename = ", filename_split )
    
    basepath = []
    for part in filename_split:
        
        if part == "maps":
            break
        
        basepath = os.path.join( basepath, part )
    
    print( "basepath = ", basepath )
    
    deffilename = os.path.join( basepath, 'exported', 'entities.json' )
    
    print( "definitions file = ", deffilename )
    
    defs = json.loads( open( deffilename ).read() )

    #if "gravity" in defs["entities"]["aas48"]: 
    #    print( "cool aas48.gravity = {0}".format( data["entities"]["aas48"]["gravity"] ) )
    
    start = time.time() 
    data = json.loads( open( filename ).read() )
    end = time.time()
    print( "loading {0} took {1} seconds".format( filename, ( end - start ) ) )
    
    numEntities = len( data["entities"] )
    for entNum in range( numEntities ):
        
        ent = data["entities"][entNum]
     
        origin = ( 0, 0, 0 )
        entname = "worldspawn"
        classname = ent["classname"]
        
        if "origin" in ent:
            origin = get_vector( ent, "origin" )
            
        if "name" in ent:
            entname = ent["name"]
            
        print( "creating entity {0} of {1}: {2}".format( entNum, numEntities, entname ) )
            
        #group = bpy.data.groups.new( entname )
        #group = bpy.ops.object.empty_add()
        
        entObj = None
        
        if "light_radius" in ent:
            # create new lamp datablock
            lamp_data = bpy.data.lamps.new( name = entname, type='POINT')
            
            """
            FIXME FIXME
            """
            radius = get_vector( ent, "light_radius" )
            print( radius.magnitude )
            lamp_data.distance = radius.length
            lamp_data.use_nodes = True

            # create new object with our lamp datablock
            entObj = bpy.data.objects.new( name = entname, object_data=lamp_data )

            # link lamp object to the scene so it'll appear in this scene
            scene = bpy.context.scene
            scene.objects.link(entObj)

            # place lamp to a specified location
            entObj.location = origin
            entObj.show_name = True

            # and finally select it make active
            entObj.select = True
            scene.objects.active = entObj
            
        
        else:
        
            bpy.ops.object.add(
                type='EMPTY', 
                enter_editmode=False,
                location=origin)
        
            entObj = bpy.context.object
            entObj.name = entname
            entObj.show_name = True
            entObj.select = True
            
        bpy.context.scene.objects.active = entObj
        
        for key in ent:
            
            if key != "primitives":
                
                value = ent[ key ]
                
                if key in defs["entities"][classname]:
                    #print( "game property {0}.{1} = {2}, default = {3}".format( classname, key, value, defs["entities"][classname][key] ) )
                    #group[ key ] = ent[ key ]
                    
                    #bpy.ops.object.game_property_new( type = 'STRING', name = key )
                    
                    if key != "entity" and key != "name" and key != "origin" and key != "angle":
                        
                        if key == "nodiffuse":
                            entObj[ key ]= int( value )
                        else:
                            entObj[ key ] = value
                else:
                    #if key != "entity" and key != "name" and key != "origin" and key != "angle" and key != "rotation":
                    #    print( "unknown game property {0}.{1} = {2}".format( classname, key, value ) )
                    
                    #bpy.ops.object.game_property_new( type = 'STRING', name = key )
                    #bpy.context.object.value = value
                    
                    if key != "entity" and key != "name" and key != "origin" and key != "angle":
                        entObj[ key ] = value
        
        #print( "origin =", origin )
     
        if "primitives" in ent:
            for i in range( len( ent["primitives"] ) ):
                
                prim = ent["primitives"][i]
                #print( "numverts =", len( prim["verts"] ) )
                
                primName = "{0}.prim{1}".format( entname, i )
                
                me = bpy.data.meshes.new( primName )
                ob = bpy.data.objects.new( primName, me )
                #ob.location = origin
                ob.show_name = True
     
                # link object to scene and make active
                scn = bpy.context.scene
                scn.objects.link(ob)
                scn.objects.active = ob
                ob.select = True
        
                #bpy.ops.object.group_link( group = group.name )
                ob.parent = entObj
     
                
                # build basic mesh
                verts = []
                for v in prim["verts"]:
                    verts.append( ( v["xyz"][0], v["xyz"][1], v["xyz"][2] ) )
                    
                faces = []
                for f in prim["polygons"]:
                    faces.append( f["indices"] )
                    
                me.from_pydata(verts, [], faces)
                
                # assign UVs
                me.uv_textures.new()
                for i in range( len( prim["verts"] ) ):
                    v = prim["verts"][i]
                    me.uv_layers[0].data[i].uv = ( v["st"][0], 1.0 - v["st"][1] )
                    
                # collect materials
                materials = []
                for f in prim["polygons"]:
                    mat = f["material"]
                    if not mat in materials:
                        materials.append( mat )
                
                # create blender materials 
                for mat in materials:
                    #print( mat )
                    
                    bmat = None
                    if mat in bpy.data.materials:
                        bmat = bpy.data.materials[ mat ]
                    else:
                        bmat = bpy.data.materials.new( mat )
                    
                    me.materials.append( bmat )
                    
                # create blender groups for special materials like textures/common/clip or visportals
                groupDict = [
                    ( "textures/common/caulk", "caulk" ),
                    ( "textures/common/nodraw", "nodraw" ),
                    ( "textures/common/nodrawsolid", "nodrawsolid" ),
                    
                    ( "textures/common/collision", "collision" ),
                    ( "textures/common/clip", "clip" ),
                    ( "textures/common/player_clip", "player_clip" ),
                    ( "textures/common/player_clip", "clip" ),
                    ( "textures/common/full_clip", "full_clip" ),
                    ( "textures/common/full_clip", "clip" ),
                    ( "textures/common/monster_clip", "monster_clip" ),
                    ( "textures/common/monster_clip", "clip" ),
                    ( "textures/common/ik_clip", "ik_clip" ),
                    ( "textures/common/ik_clip", "clip" ),
                    ( "textures/common/movable_clip", "movable_clip" ),
                    ( "textures/common/movable_clip", "clip" ),
                    ( "textures/common/clip_plusmovables", "clip_plusmovables" ),
                    ( "textures/common/clip_plusmovables", "clip" ),
                    
                    ( "textures/common/cushion", "cushion" ),
                    ( "textures/common/slick", "slick" ),
                    ( "textures/common/noimpact", "noimpact" ),
                    ( "textures/common/nodrop", "nodrop" ),
                    ( "textures/common/ladder", "ladder" ),
                    
                    ( "textures/common/mirror", "mirror" ),
                    
                    ( "textures/common/shadow", "shadow" ),
                    ( "textures/common/shadow2", "shadow2" ),
                    
                    ( "textures/common/trigger", "trigger" ),
                    ( "textures/common/trigmulti", "trigger" ),
                    ( "textures/common/trigonce", "trigger" ),
                    ( "textures/common/trigtimer", "trigger" ),
                    ( "textures/common/trigrelay", "trigger" ),
                    ( "textures/common/trighurt", "trigger" ),
                    ( "textures/common/trigfade", "trigger" ),
                    ( "textures/common/triggui", "trigger" ),
                    ( "textures/common/trigentityname", "trigger" ),
                    ( "textures/common/trigentityname_once", "trigger" ),
                    
                    ( "textures/editor/entitygui", "gui" ),
                    ( "textures/editor/entitygui2", "gui" ),
                    ( "textures/editor/entitygui3", "gui" ),
                    ( "textures/editor/pda_gui", "gui" ),
                    
                    ( "textures/editor/aassolid", "aassolid" ),
                    ( "textures/editor/aassolid", "aas" ),
                    ( "textures/editor/aasobstacle", "aasobstacle" ),
                    ( "textures/editor/aassolid", "aas" ),
                    
                    ( "textures/editor/visportal", "visportal" )
                ]
                
                for mat in materials:
                    
                    for ( tex, groupName ) in groupDict:
                        if mat == tex:
                            group = None
                            if groupName in bpy.data.groups:
                                group = bpy.data.groups[ groupName ]
                                group.objects.link( ob )
                            else:
                                bpy.ops.group.create( name = groupName )
                            
                
                    
                for i in range( len( prim["polygons"] ) ):
                    f = prim["polygons"][i]
                    mat = f["material"]
                    me.polygons[i].material_index = materials.index( mat )
                    #print( "polygon {0} material_index = {1}".format( i, me.polygons[i].material_index ) )
                
                me.update()
                
                ob.select = False
            
        # deselect so we don't add objects into the wrong groups
        entObj.select = False



class ImportMap( bpy.types.Operator, ImportHelper ):
    """Load a JSON file"""
    bl_idname = "import_scene.rbdoom_map_json"
    bl_label = "Import RBDOOM-3-BFG JSON map"
    bl_options = {'UNDO', 'PRESET'}

    directory = StringProperty()

    filename_ext = ".json"
    filter_glob = StringProperty( default="*.json", options = { 'HIDDEN' } )

    def execute(self, context):
        import_map( self.filepath )
        
        return { 'FINISHED' }
        

def menu_func_import(self, context):
    self.layout.operator( ImportMap.bl_idname, text="RBDOOM-3-BFG Map (.json)")


def register():
    bpy.utils.register_module(__name__)

    bpy.types.INFO_MT_file_import.append(menu_func_import)


def unregister():
    bpy.utils.unregister_module(__name__)

    bpy.types.INFO_MT_file_import.remove(menu_func_import)

if __name__ == "__main__":
    register()


# THIS SCRIPT IS EXPERIMENTAL, UNFINISHED AND UNSUPPORTED

# RB: My intention of this script is to provide it as a backup before it gets lost.
# Feel free to experiment with it and to load the id Tech 4.x Flash guis into Blender without animation

import json, sys, bpy
import mathutils
import time
import os
from decimal import *
from math import *

#jsonfilename = "C:\\Projects\\RBDOOM-3-BFG\\base\\exported\\swf\\shell.json"
jsonfilename = "C:\\Projects\\RBDOOM-3-BFG\\mod_ragetoolkit\\exported\\swf\\hud.json"

basepath = "C:\\Projects\\RBDOOM-3-BFG\\mod_ragetoolkit\\"

start = time.time() 
data = json.loads( open( jsonfilename ).read() )
end = time.time()
print( "loading {0} took {1} seconds".format( jsonfilename, ( end - start ) ) )


scene = bpy.context.scene

#[ %f, %f, %f, %f, %f, %f ]", m.xx, m.yy, m.xy, m.yx, m.tx, m.ty
def transform_by_stylematrix( m, uv ):
    return ( ( uv[0] * m[0] ) + ( uv[1] * m[2] ) + m[4], ( uv[1] * m[1] ) + ( uv[0] * m[3] ) + m[5] )

def inverse_stylematrix( m ):
    inverse = [ 0, 0, 0, 0, 0, 0 ];
      
    det = ( ( m[0] * m[1] ) - ( m[2] * m[3] ) )
	#if( idMath::Fabs( det ) < idMath::FLT_SMALLEST_NON_DENORMAL )
	#{
	#	return *this;
	#}
    
    invDet = 1.0 / det
    
    inverse[0] = invDet * m[1]
    inverse[3] = invDet * -m[3]
    inverse[2] = invDet * -m[2]
    inverse[1] = invDet * m[0]
    #inverse.tx = invDet * ( xy * ty ) - ( yy * tx );
    #inverse.ty = invDet * ( yx * tx ) - ( xx * ty );
    return inverse


def convert_flash_object( entry, buildDict ):
    #print( entry["type"] )
    
    origin = ( 0, 0, 0 )
    characterID = entry["characterID"]
    characterIDStr = "characterID.{0}".format( entry["characterID"] )
    
    #print( "processing characterID {0} = {1}".format( characterID, entry["type"] ) )
    
    
    if entry["type"] == "IMAGE" and buildDict == True:
        
        imgfile = entry["imageFile"]
        imgfilename = os.path.normpath( basepath + imgfile )
        
        width = entry["width"]
        height = entry["height"]
        
        if os.path.exists( imgfilename ):
            print( "found " + imgfilename )
            
        # create material
        mat = None
        if characterIDStr in bpy.data.materials:
            mat = bpy.data.materials[ characterIDStr ]
        else:
            mat = bpy.data.materials.new( characterIDStr )
            mat.preview_render_type = 'FLAT'
            mat.use_nodes = True
            
            matNodes = mat.node_tree.nodes
            matLinks = mat.node_tree.links
            
            matDisney = matNodes.get( "Principled BSDF" )
            
            #bmattex.use_map_alpha = True
        
            # create image and assign to material
            image = None
            if characterIDStr in bpy.data.images:
                image = bpy.data.images[ characterIDStr ]
            else:
                image = bpy.data.images.load( imgfilename )
        
            #btex = None
            #if characterIDStr in bpy.data.textures:
            #    btex = bpy.data.textures[ characterIDStr ]
            #else:
            #    btex = bpy.data.textures.new( characterIDStr, type = 'IMAGE' )
            #btex.image = image
            
            # create image node
            nodeTex = matNodes.new( "ShaderNodeTexImage" )
            nodeTex.image = image
            
            # create the texture coordinate
            nodeUV = matNodes.new( "ShaderNodeTexCoord" )
            
            matLinks.new( nodeTex.inputs["Vector"], nodeUV.outputs["UV"] )
            matLinks.new( matDisney.inputs["Base Color"], nodeTex.outputs["Color"] )
            matLinks.new( matDisney.inputs["Alpha"], nodeTex.outputs["Alpha"] )
            
            mat.blend_method = 'BLEND'
        
    
    if entry["type"] == "SHAPE" and buildDict == True:
        
        if "startBounds" in entry:
            startBounds = entry["startBounds"]
            origin = ( startBounds[0], startBounds[1], 0 )
        
        meshName = "shape.{0}.mesh".format( characterID )
        
        mesh = bpy.data.meshes.new( meshName )
        meshObj = bpy.data.objects.new( meshName, mesh )
        meshObj.location = origin
        #ob.show_name = True
        
        # give object unique characterID so any sprite can reference it
        meshObj["characterID"] = characterID
 
        # link object to scene and make active
        #bpy.context.collection.objects.link( meshObj )
        #bpy.context.scene.objects.active = ob
        #ob.select = True
        
        #shapeName = "shape.{0}".format( characterID )
        shapeCollection = bpy.data.collections.new( characterIDStr )
        shapeCollection.color_tag = 'COLOR_04'
        bpy.context.scene.collection.children.link( shapeCollection )
        shapeCollection["characterID"] = characterID
        
        shapeCollection.objects.link( meshObj )
        bpy.data.collections[ "Dictionary" ].children.link( shapeCollection )
        
        # add to SWF dictionary
        #bpy.data.collections[ "Dictionary" ].objects.link( meshObj )
        
        # remove from scene default collection
        bpy.context.scene.collection.children.unlink( shapeCollection )
        #bpy.context.scene.collection.objects.unlink( meshObj )
        
        #bpy.ops.collection.create( name = characterIDStr )
        #group = bpy.data.collections[ characterIDStr ]
        
        
        #for i in range( len( entry["fillDraws"] ) ):   
        #    fillDraw = entry["fillDraws"][i]
        
        #if len( entry["fillDraws"] ) > 1:
        #    error
        
        # build basic mesh from all fillDraws and assign different materials to the faces
        verts = []
        faces = []
        numverts = 0
        
        if "fillDraws" in entry:
            for fillDraw in entry["fillDraws"]:
            
                if "startVerts" not in fillDraw:
                    continue
            
                # convert triangles to faces
                indices = fillDraw["indices"]
                for i in range( 0, len( indices ), 3 ):
                    faces.append( ( numverts + indices[i], numverts + indices[i+1], numverts + indices[i+2] ) )
                
                fillDraw["firstVert"] = numverts
                
                for v in fillDraw["startVerts"]:
                    verts.append( ( v["v"][0], v["v"][1], 0.0 ) )
                    numverts += 1
                    
        if "lineDraws" in entry:
            
            print( "characterID {0} = {1} has lineDraws".format( characterID, entry["type"] ) )
            
            for lineDraw in entry["lineDraws"]:
            
                if "startVerts" not in lineDraw:
                    continue
            
                # convert triangles to faces
                indices = lineDraw["indices"]
                for i in range( 0, len( indices ), 3 ):
                    faces.append( ( numverts + indices[i], numverts + indices[i+1], numverts + indices[i+2] ) )
                
                lineDraw["firstVert"] = numverts
                
                for v in lineDraw["startVerts"]:
                    verts.append( ( v["v"][0], v["v"][1], 0.0 ) )
                    numverts += 1
                
        mesh.from_pydata( verts, [], faces )
        
        # set mesh object to active
        bpy.context.view_layer.objects.active = meshObj
        meshObj.select_set( True )
        
        # convert tris to quads
        bpy.ops.object.mode_set(mode = 'EDIT')
        bpy.ops.mesh.tris_convert_to_quads()
        bpy.ops.object.editmode_toggle()
        
        # create uv map
        mesh.uv_layers.new( do_init = False, name="UVMap")
        #mesh.update()
        
        if "fillDraws" in entry:
            #for fillDraw in entry["fillDraws"]:
            for i in range( len( entry["fillDraws"] ) ):
                
                fillDraw = entry["fillDraws"][i]
                
                if fillDraw["style"]["type"] == "bitmap":
                    
                    bitmapID = fillDraw["style"]["bitmapID"]
                    
                    if bitmapID == 65535:
                        continue   
                    
                    stylematrix = inverse_stylematrix( fillDraw["style"]["startMatrix"] )
                   
                    # build uv coords
                    #firstVert = fillDraw["firstVert"]
                    
                    # FIXME only calculate UVs vertices of this bitmap
                    for i in range( len( mesh.vertices ) ):
                        v = mesh.vertices[i]
                        
                        width = startBounds[2] - startBounds[0]
                        height = startBounds[3] - startBounds[1]
                        
                        uv = ( ( ( v.co[0] - startBounds[0] ) * ( 1.0 / width ) * 1.0 ) , ( v.co[1] - startBounds[1] ) * ( 1.0 / height ) * 1.0 )
                        uv = ( 1.0 - uv[0], uv[1] )
                        mesh.uv_layers[0].data[i].uv = uv
                        
                        #uv = ( v.co[0] * ( 1.0 / startBounds["width"] ) * 20.0 , v.co[1] * ( 1.0 / startBounds["height"] ) * 20.0 )
                        #uv = ( 1.0 - uv[0], uv[1] )
                        #me.uv_layers[0].data[i].uv = transform_by_stylematrix( stylematrix, uv )
                        
                        if characterID == 135:
                            print( "v = ({0},{1}) uv = ({2},{3})".format( v.co[0], v.co[1], uv[0], uv[1] ) )
                    
                    
                    # assign bitmap
                    
                    bitmap = "characterID.{0}".format( bitmapID )
                    
                    if bitmap not in mesh.materials:
                        bitmapmat = bpy.data.materials[ "characterID.{0}".format( bitmapID ) ]
                        mesh.materials.append( bitmapmat )
                    
                    #mesh.update()
                    
                    
                if fillDraw["style"]["type"] == "solid":
                    
                    startColor = [ 1.0, 1.0, 1.0, 1.0 ]
                    
                    if "startColor" in fillDraw["style"]:
                        startColor = fillDraw["style"]["startColor"]
                      
                    solidmat = bpy.data.materials.new( "shape.{0}.fillDraw.{1}".format( characterID, i ) )
                    solidmat.preview_render_type = 'FLAT'
                    #solidmat.use_shadeless = True
                    #solidmat.use_transparency = True
                    solidmat.diffuse_color = ( startColor[0], startColor[1], startColor[2], startColor[3] )
                    mesh.materials.append( solidmat )
    
       
    
    if entry["type"] == "SPRITE" and buildDict == False:
        
        #spriteObj = None
        
        bpy.ops.object.add(
            type='EMPTY', 
            enter_editmode=False,
            location=origin)
            
        spriteObj = bpy.context.object
        
        if "mainsprite" in entry:
            
            spriteObj.name = "mainsprite"
            spriteObj.show_name = True
        else:
            spriteObj.name = "sprite.{0}".format( characterID )
        
        spriteObj["characterID"] = characterID
        
        
        spriteCollection = bpy.data.collections.new( characterIDStr )
        spriteCollection.color_tag = 'COLOR_03'
        bpy.context.scene.collection.children.link( spriteCollection )
        
        #spriteCollection = bpy.data.collections[ characterIDStr ]
        spriteCollection["characterID"] = characterID
        spriteCollection.objects.link( spriteObj )
        
        # add to SWF Sprites
        bpy.data.collections[ "Sprites" ].children.link( spriteCollection )
                        
        # remove from scene default collection
        bpy.context.scene.collection.children.unlink( spriteCollection )
        #bpy.context.scene.collection.objects.unlink( placeObj )
        
        for command in entry["commands"]:
            
            if command["type"] == "Tag_PlaceObject2" or command["type"] == "Tag_PlaceObject3":
                
                if "characterID" in command:
                    sourceID = command["characterID"]
                    
                    bpy.ops.object.select_all( action = 'DESELECT' )
                    
                    #print( bpy.context.selected_objects )
                    
                    #print( "searching sourceID ", sourceID )
                    
                    sourceCollection = None
                    for source in bpy.data.collections:
                        if "characterID" in source and source["characterID"] == sourceID:
                            sourceCollection = source
                            break
                    
                    if sourceCollection == None:
                        print( "missed clone source ", sourceID )
                    
                    else:
                        
                        #print( "duplicating target = ", target.name )
                        #print( bpy.context.selected_objects )
                        
                        # place instance at startMatrix
                        m = [1.0, 1.0, 0.0, 0.0, 0.0, 0.0 ]
                        
                        if "startMatrix" in command:
                            m = command["startMatrix"]
                            
                        #targetLocation = ( m[4], m[5], command["depth"] )
                        #targetLocation = spriteObj.location + mathutils.Vector( ( m[4], m[5], -1.0 ) )
                        
                        #targetScale = ( m[0], m[1], 1.0 )
                        
                        # https://blender.stackexchange.com/questions/156473/how-to-avoid-operator-collection-instance-add
                        
                        #bpy.ops.object.duplicate()
                        placeObj = bpy.data.objects.new( name = sourceCollection.name, object_data = None )
                        placeObj.instance_collection = sourceCollection
                        placeObj.instance_type = 'COLLECTION'
                        
                        parentCollection = spriteCollection
                        parentCollection.objects.link( placeObj )
                        
                        #parentCollection = bpy.context.view_layer.active_layer_collection
                        #parentCollection.collection.objects.link( placeObj )
                        
                        
                        
                        #if "mainsprite" in entry:
                        #    placeObj.name = "mainsprite"
                        #else:
                        #    placeObj.name = "sprite.{0}".format( characterID )
                            
                        if "name" in command:
                            placeObj.name = "{0}.{1}.{2}".format( spriteObj.name, command["name"], sourceID )
                            placeObj.show_name = True
                        else:
                            placeObj.name = "{0}.characterID.{1}".format( spriteObj.name, sourceID )
                            
                        placeObj.parent = spriteObj
                        #targetClone["characterID"] = -1
                        
                        placeObj.location = ( m[4], m[5], command["depth"] )
                        placeObj.scale = ( m[0], m[1], 1.0 )
                                       
                        bpy.ops.object.select_all( action = 'DESELECT' )
                        
                           

def create_default_collections():
    
    # remove all groups
    for group in bpy.data.collections:
        bpy.data.collections.remove( group )
        
    #bpy.ops.collection.create( name = "Dictionary" )
    #bpy.ops.collection.create( name = "Sprites" )
    collection = bpy.data.collections.new( "Dictionary" )
    bpy.context.scene.collection.children.link( collection )
    
    collection = bpy.data.collections.new( "Sprites" )
    bpy.context.scene.collection.children.link( collection )
    
    print( bpy.data.collections )
    

# clear old materials
for mat in bpy.data.materials:
    mat.user_clear()
    bpy.data.materials.remove( mat )

create_default_collections()

start = time.time() 

for entry in data["dict"]:
    convert_flash_object( entry, True )

i = 0
for entry in data["dict"]:
    convert_flash_object( entry, False )
    
    i += 1
    
    #if i == 4:
    #   break
    
end = time.time()
print( "importing {0} took {1} seconds".format( jsonfilename, ( end - start ) ) )


# THIS SCRIPT IS EXPERIMENTAL, UNFINISHED AND UNSUPPORTED

# RB: My intention of this script is to provide it as a backup before it gets lost.
# Feel free to experiment with it and to load the id Tech 4.x Flash guis into Blender without animation

import json, sys, bpy
import mathutils
import time
import os
from decimal import *
from math import *

jsonfilename = "G:\\Projects\\RBDOOM-3-BFG\\base\\exported\\swf\\shell.json"
basepath = "G:\\Projects\\RBDOOM-3-BFG\\base\\"

start = time.time() 
data = json.loads( open( jsonfilename ).read() )
end = time.time()
print( "loading {0} took {1} seconds".format( jsonfilename, ( end - start ) ) )


scene = bpy.context.scene

#[ %f, %f, %f, %f, %f, %f ]", m.xx, m.yy, m.xy, m.yx, m.tx, m.ty
def transform_by_stylematrix( m, uv ):
    return ( ( uv[0] * m[0] ) + ( uv[1] * m[2] ) + m[4], ( uv[1] * m[1] ) + ( uv[0] * m[3] ) + m[5] )

def inverse_stylematrix( m ):
    inverse = [ 0, 0, 0, 0, 0, 0 ];
      
    det = ( ( m[0] * m[1] ) - ( m[2] * m[3] ) )
	#if( idMath::Fabs( det ) < idMath::FLT_SMALLEST_NON_DENORMAL )
	#{
	#	return *this;
	#}
    
    invDet = 1.0 / det
    
    inverse[0] = invDet * m[1]
    inverse[3] = invDet * -m[3]
    inverse[2] = invDet * -m[2]
    inverse[1] = invDet * m[0]
    #inverse.tx = invDet * ( xy * ty ) - ( yy * tx );
    #inverse.ty = invDet * ( yx * tx ) - ( xx * ty );
    return inverse


def convert_flash_object( entry ):
    #print( entry["type"] )
    
    origin = ( 0, 0, 0 )
    characterID = entry["characterID"]
    characterIDStr = "characterID.{0}".format( entry["characterID"] )
    
    #print( "processing characterID {0} = {1}".format( characterID, entry["type"] ) )
    
    
    if entry["type"] == "IMAGE":
        
        imgfile = entry["imageFile"]
        imgfilename = os.path.normpath( basepath + imgfile )
        
        width = entry["width"]
        height = entry["height"]
        
        if os.path.exists( imgfilename ):
            print( "found " + imgfilename )
            
        # create material
        bmat = None
        if characterIDStr in bpy.data.materials:
            bmat = bpy.data.materials[ characterIDStr ]
        else:
            bmat = bpy.data.materials.new( characterIDStr )
            bmat.preview_render_type = 'FLAT'
            bmat.use_shadeless = True
            bmat.use_transparency = True
            bmat.alpha = 0
            bmattex = bmat.texture_slots.add()
            bmattex.use_map_alpha = True
        
            # create image and assign to material
            image = None
            if characterIDStr in bpy.data.images:
                image = bpy.data.images[ characterIDStr ]
            else:
                image = bpy.data.images.load( imgfilename )
            image.use_alpha = True
        
            btex = None
            if characterIDStr in bpy.data.textures:
                btex = bpy.data.textures[ characterIDStr ]
            else:
                btex = bpy.data.textures.new( characterIDStr, type = 'IMAGE' )
            btex.image = image
            
            bmattex.texture = btex
            bmattex.texture_coords = 'UV'
            bmattex.mapping = 'FLAT'
        
    
    if entry["type"] == "SHAPE":
        
        if "startBounds" in entry:
            startBounds = entry["startBounds"]
            origin = ( startBounds[0], startBounds[1], 0 )
        
        meName = "shape.{0}".format( characterID )
        
        me = bpy.data.meshes.new( meName )
        ob = bpy.data.objects.new( meName, me )
        ob.location = origin
        #ob.show_name = True
        
        # give object unique characterID so any sprite can reference it
        ob["characterID"] = characterID
 
        # link object to scene and make active
        scene.objects.link( ob )
        scene.objects.active = ob
        ob.select = True
        
        bpy.ops.group.create( name = characterIDStr )
        group = bpy.data.groups[ characterIDStr ]
        group["characterID"] = characterID
        
        #for i in range( len( entry["fillDraws"] ) ):   
        #    fillDraw = entry["fillDraws"][i]
        
        #if len( entry["fillDraws"] ) > 1:
        #    error
        
        # build basic mesh from all fillDraws and assign different materials to the faces
        verts = []
        faces = []
        numverts = 0
        
        if "fillDraws" in entry:
            for fillDraw in entry["fillDraws"]:
            
                if "startVerts" not in fillDraw:
                    continue
            
                # convert triangles to faces
                indices = fillDraw["indices"]
                for i in range( 0, len( indices ), 3 ):
                    faces.append( ( numverts + indices[i], numverts + indices[i+1], numverts + indices[i+2] ) )
                
                fillDraw["firstVert"] = numverts
                
                for v in fillDraw["startVerts"]:
                    verts.append( ( v["v"][0], v["v"][1], 0.0 ) )
                    numverts += 1
                    
        if "lineDraws" in entry:
            
            print( "characterID {0} = {1} has lineDraws".format( characterID, entry["type"] ) )
            
            for lineDraw in entry["lineDraws"]:
            
                if "startVerts" not in lineDraw:
                    continue
            
                # convert triangles to faces
                indices = lineDraw["indices"]
                for i in range( 0, len( indices ), 3 ):
                    faces.append( ( numverts + indices[i], numverts + indices[i+1], numverts + indices[i+2] ) )
                
                lineDraw["firstVert"] = numverts
                
                for v in lineDraw["startVerts"]:
                    verts.append( ( v["v"][0], v["v"][1], 0.0 ) )
                    numverts += 1
                
        me.from_pydata( verts, [], faces )
        
        bpy.ops.object.editmode_toggle()
        #bpy.ops.mesh.quads_convert_to_tris( quad_method = 'BEAUTY', ngon_method = 'BEAUTY' )
        bpy.ops.mesh.tris_convert_to_quads()
        #bpy.ops.uv.reset()
        #bpy.ops.uv.unwrap()
        bpy.ops.object.editmode_toggle()
        
        # create uv map
        me.uv_textures.new()
        me.update()
        
        if "fillDraws" in entry:
            #for fillDraw in entry["fillDraws"]:
            for i in range( len( entry["fillDraws"] ) ):
                
                fillDraw = entry["fillDraws"][i]
                
                if fillDraw["style"]["type"] == "bitmap":
                    
                    bitmapID = fillDraw["style"]["bitmapID"]
                    
                    if bitmapID == 65535:
                        continue   
                    
                    stylematrix = inverse_stylematrix( fillDraw["style"]["startMatrix"] )
                   
                    # build uv coords
                    #firstVert = fillDraw["firstVert"]
                    
                    # FIXME only calculate UVs vertices of this bitmap
                    for i in range( len( me.vertices ) ):
                        v = me.vertices[i]
                        
                        width = startBounds[2] - startBounds[0]
                        height = startBounds[3] - startBounds[1]
                        
                        uv = ( ( ( v.co[0] - startBounds[0] ) * ( 1.0 / width ) * 1.0 ) , ( v.co[1] - startBounds[1] ) * ( 1.0 / height ) * 1.0 )
                        uv = ( 1.0 - uv[0], uv[1] )
                        me.uv_layers[0].data[i].uv = uv
                        
                        #uv = ( v.co[0] * ( 1.0 / startBounds["width"] ) * 20.0 , v.co[1] * ( 1.0 / startBounds["height"] ) * 20.0 )
                        #uv = ( 1.0 - uv[0], uv[1] )
                        #me.uv_layers[0].data[i].uv = transform_by_stylematrix( stylematrix, uv )
                        
                        if characterID == 135:
                            print( "v = ({0},{1}) uv = ({2},{3})".format( v.co[0], v.co[1], uv[0], uv[1] ) )
                    
                    
                    # assign bitmap
                    
                    bitmap = "characterID.{0}".format( bitmapID )
                    
                    if bitmap not in me.materials:
                        bitmapmat = bpy.data.materials[ "characterID.{0}".format( bitmapID ) ]
                        me.materials.append( bitmapmat )
                    
                    me.update()
                    
                    
                if fillDraw["style"]["type"] == "solid":
                    
                    startColor = [ 1.0, 1.0, 1.0, 1.0 ]
                    
                    if "startColor" in fillDraw["style"]:
                        startColor = fillDraw["style"]["startColor"]
                      
                    solidmat = bpy.data.materials.new( "shape.{0}.fillDraw.{1}".format( characterID, i ) )
                    solidmat.preview_render_type = 'FLAT'
                    solidmat.use_shadeless = True
                    solidmat.use_transparency = True
                    solidmat.diffuse_color = ( startColor[0], startColor[1], startColor[2] )
                    solidmat.alpha = startColor[3]
                    me.materials.append( solidmat )
    
       
    
    if entry["type"] == "SPRITE":
        
        bpy.ops.object.add(
            type='EMPTY', 
            enter_editmode=False,
            location=origin)
            
        sprite = bpy.context.object
        
        if "mainsprite" in entry:
            sprite.name = "mainsprite"
            sprite.show_name = True
        else:
            sprite.name = "sprite.{0}".format( characterID )
        
        sprite["characterID"] = characterID
        
        bpy.ops.group.create( name = characterIDStr )
        
        spriteGroup = bpy.data.groups[ characterIDStr ]
        spriteGroup["characterID"] = characterID
        
        #sprite.select = True
                
        #scene.objects.active = ob
        
        for command in entry["commands"]:
            
            if command["type"] == "Tag_PlaceObject2":
                
                if "characterID" in command:
                    targetID = command["characterID"]
                    
                    # instantiate target
                    bpy.ops.object.select_all( action = 'DESELECT' )
                    
                    #print( bpy.context.selected_objects )
                    
                    #print( "searching targetID ", targetID )
                    
                    target = None
                    for group in bpy.data.groups:
                        if "characterID" in group and group["characterID"] == targetID:
                            target = group
                            break
                    
                    if target == None:
                        print( "missed target ", targetID )
                    
                    
                    if target != None:
                        
                        #print( "duplicating target = ", target.name )
                        #print( bpy.context.selected_objects )
                        
                        #bpy.ops.object.duplicate()
                        bpy.ops.object.group_instance_add( name = target.name )
                        bpy.ops.object.group_link( group = spriteGroup.name )
                        
                        targetClone = bpy.context.selected_objects[0]
                        
                        if "name" in command:
                            targetClone.name = "{0}.{1}".format( sprite.name, command["name"] )
                            targetClone.show_name = True
                        else:
                            targetClone.name = "{0}.characterID.{1}".format( sprite.name, targetID )
                            
                        targetClone.parent = sprite
                        #targetClone["characterID"] = -1
                        
                        # move instance to startMatrix
                        m = [1.0, 1.0, 0.0, 0.0, 0.0, 0.0 ]
                        
                        if "startMatrix" in command:
                            m = command["startMatrix"]
                            
                        targetClone.location = ( m[4], m[5], -1.0 )
                        targetClone.scale = ( m[0], m[1], 1.0 )
                        
                        bpy.ops.object.select_all( action = 'DESELECT' )
                        
                           
    
    """
    if entry["type"] == "SPRITE":
        
        bpy.ops.object.add(
            type='EMPTY', 
            enter_editmode=False,
            location=origin)
            
        sprite = bpy.context.object
        sprite.name = "sprite.{0}".format( characterID )
        #sprite.show_name = True
        sprite["characterID"] = characterID
        #sprite.select = True
                
        #scene.objects.active = ob
        
        for command in entry["commands"]:
            
            if command["type"] == "Tag_PlaceObject2":
                
                if "characterID" in command:
                    targetID = command["characterID"]
                    
                    # instantiate target
                    bpy.ops.object.select_all( action = 'DESELECT' )
                    
                    #print( bpy.context.selected_objects )
                    
                    #print( "searching targetID ", targetID )
                    
                    target = None
                    for obj in scene.objects:
                        if "characterID" in obj and obj["characterID"] == targetID:
                            target = obj
                            break
                    
                    if target == None:
                        print( "missed target ", targetID )
                    
                    
                    if target != None:
                        target.select = True
                        
                        print( "duplicating target = ", target.name )
                        print( bpy.context.selected_objects )
                        
                        bpy.ops.object.duplicate()
                        
                        targetClone = bpy.context.selected_objects[0]
                        
                        if "name" in command:
                            targetClone.name = "{0}.{1}".format( sprite.name, command["name"] )
                        else:
                            targetClone.name = "{0}.characterID.{1}".format( sprite.name, targetID )
                            
                        targetClone.parent = sprite
                        targetClone["characterID"] = -1
                        
                        bpy.ops.object.select_all( action = 'DESELECT' )
                        
                        for obj in target.children:
                            obj.select = True
                        
                        #scene.objects.active.parent = target
                        #bpy.ops.object.select_grouped( type = 'CHILDREN_RECURSIVE' )
                        print( bpy.context.selected_objects )
                        
                        bpy.ops.object.duplicate()
                        
                        for clone in bpy.context.selected_objects:
                            
                            if "name" in command:
                                clone.name = "{0}.{1}".format( targetClone.name, clone.name )
                            else:
                                clone.name = "{0}.{1}".format( sprite.name, clone.name )
                            clone.parent = targetClone
                            clone["characterID"] = -1
                        
                            m = command["startMatrix"]
                            clone.location = ( m[4], m[5], 0.0 )
                            clone.scale = ( m[0], m[1], 1.0 )
                        
    """          



#remove all groups
for group in bpy.data.groups:
    bpy.data.groups.remove( group )

print( bpy.data.groups )

for entry in data["dict"]:
    convert_flash_object( entry )